local L = LibStub("AceLocale-3.0"):GetLocale("NeedjaTrackerPlus", true)

NTP = LibStub("AceAddon-3.0"):NewAddon("NeedjaTrackerPlus", "AceConsole-3.0", "AceEvent-3.0")

local S2K = LibStub("S2Kmisc-1.0")

function NTP:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("NeedjaTrackerPlusDB", self:Defaults(), true)

    LibStub("AceConfig-3.0"):RegisterOptionsTable("NeedjaTrackerPlus", self:Options())

    local dlg = LibStub("AceConfigDialog-3.0")

    dlg:AddToBlizOptions("NeedjaTrackerPlus", L.title, nil, "general")
    dlg:AddToBlizOptions("NeedjaTrackerPlus", L.cfg_profiles, L.title, "profiles")

    ChatFrame_AddMessageEventFilter("CHAT_MSG_LOOT", function(this, ...) return self:FilterChatMessage(...) end)

    self:RegisterEvent("CHAT_MSG_LOOT", "ProcessChatMessage")

    self:RegisterChatCommand("nt", "ProcessChatCommand")

    self:RegisterEvent("START_LOOT_ROLL", "ProcessLootRollEvent")

    self:RegisterEvent("PLAYER_ENTERING_WORLD", "ProcessPartyEvent")
    self:RegisterEvent("LFG_COMPLETION_REWARD", "ProcessPartyEvent")
    self:RegisterEvent("RAID_ROSTER_UPDATE", "ProcessPartyEvent")
    self:RegisterEvent("PARTY_MEMBERS_CHANGED", "ProcessPartyEvent")
end

function NTP:GetProfileValue(key)
    return self.db.profile[key]
end

function NTP:SetProfileValue(key, value)
    self.db.profile[key] = value
end

function NTP:GetCharValue(key)
    return self.db.char[key]
end

function NTP:SetCharValue(key, value)
    self.db.char[key] = value
end

function NTP:GetPartyStatus()
    local status, substatus
    if GetNumRaidMembers() > 0 then
        status = "raid"
        substatus = 3
        local lmode = GetLFGMode()
        if lmode == "lfgparty" then
            status = "lfr_raid"
        end
        if IsRaidLeader() then
            substatus = 1
        elseif IsRaidOfficer() then
            substatus = 2
        end

    elseif GetNumPartyMembers() > 0 then
        status = "party"
        substatus = 2
        local lmode = GetLFGMode()
        if lmode == "lfgparty" then
            status = "lfg_party"
        end
        if IsPartyLeader() then
            substatus = 1
        end
    end
    return status, substatus
end

function NTP:FilterChatMessage(event, msg)
    return S2K:ParseMsgMatchTable(S2K.LOOT_ROLL_MSG_MATCH_TABLE, msg)
end

function NTP:ProcessChatMessage(event, msg)
    local handler, args = S2K:ParseMsgMatchTable(S2K.LOOT_ROLL_MSG_MATCH_TABLE, msg, true)

    if handler == "LOOT_ROLL_CHOICE" then
        self:ProcessRollChoice(args)

    elseif handler == "LOOT_ROLL_ROLLED" then
        self:ProcessRollRolled(args)

    elseif handler == "LOOT_ROLL_ALL_PASSED" then
        self:ProcessRollAllPassed(args)

    elseif handler == "LOOT_ROLL_WON" then
        self:ProcessRollWon(args)

    elseif handler == "LOOT_ROLL_WON_NO_SPAM" then
        self:ProcessRollWonNoSpam(args)
    end
end

function NTP:ProcessChatCommand(msg)
    local cmd, pos = self:GetArgs(msg, 1, 1)

    if cmd == "stat" or cmd == "statistics" then
        local param, channel, full
        while true do
            param, pos = self:GetArgs(msg, 1, pos)
            if param == "say" or param == "party" or param == "raid" or param == "guild" then
                channel = string.upper(param)
            elseif param == "full" then
                full = true
            end
            if pos == 1e9 then
                break
            end
        end
        self:ReportStatistics(channel, full)

    elseif cmd == "hist" or cmd == "history" then
        local param, channel, full
        while true do
            param, pos = self:GetArgs(msg, 1, pos)
            if param == "say" or param == "party" or param == "raid" or param == "guild" then
                channel = string.upper(param)
            elseif param == "full" then
                full = true
            end
            if pos == 1e9 then
                break
            end
        end
        self:ReportHistory(channel, full)

    elseif cmd == "reset" then
        self:ClearHistory()
    end
end

function NTP:ProcessLootRollEvent(event, rollid)
    local item = GetLootRollItemLink(rollid)
    self:ProcessRollStart({ item = item, rollid = rollid })
end

function NTP:ProcessPartyEvent(event)
    local clear

    local instance
    local iname, imode = GetInstanceInfo()
    if imode ~= "none" then
        instance = iname
    end
    local pinstance = self:GetCharValue("last_instance")

    if instance and pinstance ~= instance then
        self:SetCharValue("last_instance", instance)
        clear = true
    end

    if event == "PLAYER_ENTERING_WORLD" and not self.entered then
        self.entered = true
    else
        local switch = {
            none      = { none = false, party = true , lfg_party = true , raid = true , lfr_raid = true  },
            party     = { none = false, party = false, lfg_party = true , raid = true , lfr_raid = true  },
            lfg_party = { none = false, party = false, lfg_party = false, raid = false, lfr_raid = true  },
            raid      = { none = false, party = false, lfg_party = true , raid = false, lfr_raid = true  },
            lfr_raid  = { none = false, party = false, lfg_party = true , raid = false, lfr_raid = false },
        }

        local status = self:GetPartyStatus()
        local pstatus = self:GetCharValue("last_status")

        if pstatus ~= status then
            self:SetCharValue("last_status", status)
            if switch[pstatus or "none"][status or "none"] then
                clear = true
            end
        end
    end

    if clear then
        self:ClearHistory()
    end
end

function NTP:GetRollItemData(item)
    local ritem = self:GetCharValue("current")[item]
    if not ritem then
        ritem = { item = item, players = {}, need = 0, greed = 0, disenchant = 0, unknown = 0, pass = 0 }
        self:GetCharValue("current")[item] = ritem
    end
    return ritem
end

function NTP:GetAndRemoveRollItemData(item)
    local ritem = self:GetCharValue("current")[item]
    if not ritem then
        ritem = { item = item, players = {}, need = 0, greed = 0, disenchant = 0, unknown = 0, pass = 0 }
    else
        self:GetCharValue("current")[item] = nil
        ritem.rollid = nil
    end
    return ritem
end

function NTP:GetRollPlayerData(ritem, player)
    local rplayer = ritem.players[player]
    if not rplayer then
        rplayer = { player = player }
        ritem.players[player] = rplayer
    end
    return rplayer
end

function NTP:UpdateGroupLootFrame(frame, ritem)
    local aframe = _G["GroupLootFrame" .. frame:GetID() .. "Name"]
    local labels = {
        need       = { x = 0  , y = -5, r = 1.0, g = 0.2, b = 0.2, sign = "N:" },
        greed      = { x = 30 , y = -5, r = 1.0, g = 1.0, b = 0.4, sign = "G:" },
        disenchant = { x = 60 , y = -5, r = 1.0, g = 1.0, b = 0.4, sign = "D:" },
        pass       = { x = 90 , y = -5, r = 0.4, g = 1.0, b = 0.4, sign = "P:" },
        unknown    = { x = 120, y = -5, r = 1.0, g = 1.0, b = 0.4, sign = "?:" },
    }
    local choice, params
    for choice, params in pairs(labels) do
        local label = frame["NTPchoice" .. choice]
        if not label then
            label = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
            frame["NTPchoice" .. choice] = label

            label.SetTextColor(params.r, params.g, params.b)
            label.SetPoint("TOPLEFT", aframe, "BOTTOMLEFT", params.x, params.y)
        end
        label.SetText(params.sign .. ritem[choice])
    end
end

function NTP:UpdateChoiceCounters(ritem)
    ritem.need = 0 ritem.greed = 0 ritem.disenchant = 0 ritem.unknown = 0 ritem.pass = 0
    local rplayer
    for _, rplayer in pairs(ritem.players) do
        local choice = rplayer.choice or "unknown"
        ritem[choice] = ritem[choice] + 1
    end
    if ritem.rollid then
        local n
        for n = 1, NUM_LOOT_GROUP_FRAMES do
            local frame = _G["GroupLootFrame" .. n]
            if frame and frame.rollID = ritem.rollid then
                self:UpdateGroupLootFrame(frame, ritem)
                break
            end
        end
    end
end

function NTP:ProcessRollStart(args)
    local ritem = self:GetRollItemData(args.item)

    ritem.rollid = args.rollid

    self:UpdateChoiceCounters(ritem)
end

function NTP:ProcessRollChoice(args)
    local ritem = self:GetRollItemData(args.item)
    local rplayer = self:GetRollPlayerData(ritem, args.player)

    rplayer.choice = args.choice

    self:UpdateChoiceCounters(ritem)
    self:ReportRollChoice(args)
end

function NTP:ProcessRollRolled(args)
    local ritem = self:GetRollItemData(args.item)
    local rplayer = self:GetRollPlayerData(ritem, args.player)

    rplayer.choice = args.choice
    rplayer.roll = args.roll

    self:UpdateChoiceCounters(ritem)
end

function NTP:ProcessRollAllPassed(args)
    local ritem = self:GetAndRemoveRollItemData(args.item)

    self:AddToStatistics(ritem)
    self:AddToHistory(ritem)

    self:ReportRollResult(ritem)
end

function NTP:ProcessRollWon(args)
    local ritem = self:GetAndRemoveRollItemData(args.item)
    local rplayer = self:GetRollPlayerData(ritem, args.player)

    ritem.won = args.player
    ritem.choice = rplayer.choice
    ritem.roll = rplayer.roll

    self:AddToStatistics(ritem)
    self:AddToHistory(ritem)

    self:ReportRollResult(ritem)
end

function NTP:ProcessRollNoSpam(args)
    local ritem = self:GetAndRemoveRollItemData(args.item)
    local rplayer = self:GetRollPlayerData(ritem, args.player)

    ritem.won = args.player
    ritem.choice = args.choice
    ritem.roll = args.roll

    rplayer.choice = args.choice
    rplayer.roll = args.roll

    self:UpdateChoiceCounters(ritem)

    self:AddToStatistics(ritem)
    self:AddToHistory(ritem)

    self:ReportRollResult(ritem)
end

function NTP:ClearHistory()
    self:SetCharValue("stat", {})
    self:SetCharValue("history", {})
end

function NTP:AddToStatistics(ritem)
    local player, rplayer
    for player, rplayer in pairs(ritem.players) do
        local choice = rplayer.choice or "unknown"
        local splayer = self:GetCharValue("stat")[player]
        if not splayer then
            splayer = { player = player, items = {}, won = 0, need = 0, greed = 0, disenchant = 0, unknown = 0, pass = 0 }
            self:GetCharValue("stat")[player] = splayer
        end
        splayer[choice] = splayer[choice] + 1
        if ritem.won == player then
            splayer.won = splayer.won + 1
            table.insert(splayer.items, 1, ritem)
            while #splayer.items > 100 do
                table.remove(splayer.items, #splayer.items)
            end
        end
    end
end

function NTP:AddToHistory(ritem)
    local history = self:GetCharValue("history")
    table.insert(history, 1, ritem)
    while #history > 100 do
        table.remove(history, #history)
    end
end

function NTP:ReportRollChoice(args)
    if args.choice == "need" and args.player ~= UnitName("player") then
        S2K:PrintToChat((L.msg_roll_need):format(
            args.item, args.player, L["roll_" .. args.choice]
        ))

        local status, substatus = self:GetPartyStatus()
        if status then
            local warn = self:GetProfileValue("warn_" .. status)
            if warn > substatus then
                channel = "PARTY"
                if status == "raid" or status == "lfr_raid" then
                    channel = "RAID"
                end
                S2K:PrintToChat(("NTP: " .. L.msg_roll_need):format(
                    args.item, args.player, L["roll_" .. args.choice]
                ), channel)
            end
        end
    end
end

function NTP:ReportRollResult(ritem)
    if not ritem.won then
        S2K:PrintToChat((L.msg_pass):format(
            ritem.item
        ))

    elseif ritem.won == UnitName("player") then
        S2K:PrintToChat((L.msg_won_you):format(
            ritem.item,
            L["roll_" .. (ritem.choice or "unknown")],
            ritem.roll or "?"
        ))
    else
        local rplayer = ritem.players[UnitName("player")]
        if rplayer and rplayer.choice == ritem.choice then
            if ritem.choice == "need" then
                S2K:PrintToChat((L.msg_won_need_my_roll):format(
                    ritem.item, ritem.won,
                    L["roll_" .. (ritem.choice or "unknown")],
                    ritem.roll or "?", rplayer.roll or "?"
                ))
            else
                S2K:PrintToChat((L.msg_won_my_roll):format(
                    ritem.item, ritem.won,
                    L["roll_" .. (ritem.choice or "unknown")],
                    ritem.roll or "?", rplayer.roll or "?"
                ))
            end
        else
            if ritem.choice == "need" then
                S2K:PrintToChat((L.msg_won_need):format(
                    ritem.item, ritem.won,
                    L["roll_" .. (ritem.choice or "unknown")],
                    ritem.roll or "?"
                ))
            else
                S2K:PrintToChat((L.msg_won):format(
                    ritem.item, ritem.won,
                    L["roll_" .. (ritem.choice or "unknown")],
                    ritem.roll or "?"
                ))
            end
        end
    end
end

function NTP:ReportStatistics(channel, full)
    local stat = self:GetCharValue("stat")

    if #stat == 0 then
        if channel then
            S2K:PrintToChat("NTP: " .. L.msg_stat_head_empty, channel)
        else
            S2K:PrintToChat(L.msg_stat_head_empty, channel)
        end
        return
    end

    if channel then
        S2K:PrintToChat("NTP: " .. L.msg_stat_head, channel)
    else
        S2K:PrintToChat(L.msg_stat_head, channel)
    end

    local n, rplayer

    local lstat = {}
    for _, rplayer in pairs(stat) do
        table.insert(lstat, rplayer)
    end

    table.sort(lstat, function(a, b) return a.need > b.need end)
    local top = lstat[1].need
    table.sort(lstat, function(a, b) return a.won > b.won end)

    for n, rplayer in pairs(lstat) do
        if top > 0 and rplayer.need == top then
            S2K:PrintToChat((L.msg_stat_top):format(
                n, rplayer.player, rplayer.win,
                rplayer.need, rplayer.greed, rplayer.disenchant,
                rplayer.pass, rplayer.unknown
            ), channel)

        elseif rplayer.need > 0 then
            S2K:PrintToChat((L.msg_stat_need):format(
                n, rplayer.player, rplayer.win,
                rplayer.need, rplayer.greed, rplayer.disenchant,
                rplayer.pass, rplayer.unknown
            ), channel)

        else
            S2K:PrintToChat((L.msg_stat):format(
                n, rplayer.player, rplayer.win,
                rplayer.need, rplayer.greed, rplayer.disenchant,
                rplayer.pass, rplayer.unknown
            ), channel)
        end

        if not full and n >= 10 then
            break
        end
    end
end

function NTP:ReportHistory(channel, full)
    local history = self:GetCharValue("history")

    if #history == 0 then
        if channel then
            S2K:PrintToChat("NTP: " .. L.msg_hist_head_empty, channel)
        else
            S2K:PrintToChat(L.msg_hist_head_empty, channel)
        end
        return
    end

    if channel then
        S2K:PrintToChat("NTP: " .. L.msg_hist_head, channel)
    else
        S2K:PrintToChat(L.msg_hist_head, channel)
    end

    local n, ritem
    for n, ritem in paris(history) do
        if ritem.choice == "pass" then
            S2K:PrintToChat((L.msg_hist_pass):format(
                n, ritem.item
            ), channel)

        elseif ritem.choice == "need" then
            S2K:PrintToChat((L.msg_hist_need):format(
                n, ritem.item,
                ritem.player, L["roll_" .. (ritem.choice or "unknown")],
                ritem.roll or "?"
            ), channel)

        else
            S2K:PrintToChat((L.msg_hist):format(
                n, ritem.item,
                ritem.player, L["roll_" .. (ritem.choice or "unknown")],
                ritem.roll or "?"
            ), channel)
        end

        if not full and n >= 10 then
            break
        end
    end
end
